package cn.iocoder.yudao.module.wealth.job;

import cn.iocoder.yudao.framework.quartz.core.handler.JobHandler;
import cn.iocoder.yudao.module.ecw.api.bank.BankApi;
import cn.iocoder.yudao.module.ecw.api.bank.dto.BankAccountDTO;
import cn.iocoder.yudao.module.wealth.dal.dataobject.receiptItem.ReceiptItemDO;
import cn.iocoder.yudao.module.wealth.dal.dataobject.receivable.ReceivableDO;
import cn.iocoder.yudao.module.wealth.dal.dataobject.receivable.ReceivableWriteOffRecordDO;
import cn.iocoder.yudao.module.wealth.dal.mysql.receipt.ReceiptMapper;
import cn.iocoder.yudao.module.wealth.dal.mysql.receiptItem.ReceiptItemMapper;
import cn.iocoder.yudao.module.wealth.dal.mysql.receivable.ReceivableMapper;
import cn.iocoder.yudao.module.wealth.dal.mysql.receivable.ReceivableWriteOffRecordMapper;
import cn.iocoder.yudao.module.wealth.enums.ReceiptItemStatusEnum;
import cn.iocoder.yudao.module.wealth.vo.receivable.ReceivableWriteOffReqVO;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.*;
import java.util.stream.Collectors;

@Component
@Slf4j
public class UpdateReceiptItemRelationTask implements JobHandler {
    @Resource
    private ReceiptItemMapper receiptItemMapper;
    @Resource
    private ReceivableMapper receivableMapper;
    @Resource
    private BankApi bankApi;
    @Resource
    private ReceiptMapper receiptMapper;
    @Resource
    private ReceivableWriteOffRecordMapper receivableWriteOffRecordMapper;

    /**
     * 更新应收明细关联应收款
     *
     * @param param 参数
     */
    @Override
    public String execute(String param) throws Exception {
        log.info("更新应收明细关联应收款");
        long l = System.currentTimeMillis();
        List<Long> idList = receiptMapper.selectWriteOffIdList();
        LambdaQueryWrapper<ReceiptItemDO> lambdaQueryWrapper = new LambdaQueryWrapper<>();
        lambdaQueryWrapper.eq(StringUtils.isNotBlank(param), ReceiptItemDO::getReceiptId, Long.parseLong(param));
        lambdaQueryWrapper.in(StringUtils.isBlank(param), ReceiptItemDO::getId, idList);
        lambdaQueryWrapper.select(ReceiptItemDO::getReceiptId, ReceiptItemDO::getStatus, ReceiptItemDO::getId, ReceiptItemDO::getAccountId, ReceiptItemDO::getWriteOffAmount, ReceiptItemDO::getWriteOffCurrencyId, ReceiptItemDO::getCurrencyId);
        List<ReceiptItemDO> listItem = receiptItemMapper.selectList(lambdaQueryWrapper);
        // 按收款单分组
        Map<Long, List<ReceiptItemDO>> collectByReceiptId = listItem.stream().collect(Collectors.groupingBy(ReceiptItemDO::getReceiptId));
        List<ReceivableDO> receivableDOS = receivableMapper.selectList(Wrappers.lambdaQuery(ReceivableDO.class)
                .in(ReceivableDO::getReceiptId, collectByReceiptId.keySet())
                .select(ReceivableDO::getId, ReceivableDO::getReceiptId, ReceivableDO::getTaxAmount, ReceivableDO::getBaseAmount, ReceivableDO::getCurrencyId, ReceivableDO::getExchangeRate, ReceivableDO::getOrderId));
        for (Long receiptId : collectByReceiptId.keySet()) {
            List<ReceivableDO> receivableDOList = receivableDOS.stream().filter(r -> r.getReceiptId().equals(receiptId)).sorted(Comparator.comparing(ReceivableDO::getTaxAmount)).collect(Collectors.toList());
            List<ReceiptItemDO> receiptItemDOS = collectByReceiptId.get(receiptId);
            for (ReceiptItemDO receiptItemDO : receiptItemDOS) {
                boolean isWriteOff = Objects.equals(receiptItemDO.getStatus(), ReceiptItemStatusEnum.WRITE_OFF.getValue()) || Objects.equals(receiptItemDO.getStatus(), ReceiptItemStatusEnum.WRITE_OFF_NO_APPROVE_ING.getValue());
                List<ReceivableWriteOffReqVO> receiptItemReceivableWriteOff = this.getReceiptItemReceivableWriteOff(receivableDOList, receiptItemDO.getWriteOffAmount(), isWriteOff);
                for (ReceivableWriteOffReqVO receivableWriteOffReqVO : receiptItemReceivableWriteOff) {
                    this.createWriteOffRecord(receivableWriteOffReqVO, receiptItemDO);
                }
            }
            receivableMapper.updateBatch(receivableDOList);
        }
        log.info("更新应收明细关联应收款，耗时：{}", System.currentTimeMillis() - l);
        return "success";
    }

    private List<ReceivableWriteOffReqVO> getReceiptItemReceivableWriteOff(List<ReceivableDO> receivableDOList, BigDecimal writeOffAmount, boolean isWriteOff) {
        ArrayList<ReceivableWriteOffReqVO> receivableWriteOffReqVOS = new ArrayList<>();
        ReceivableDO receivableInItemVO;
        boolean isLast = false;
        for (int i = 0; i < receivableDOList.size(); i++) {
            if (i == receivableDOList.size() - 1) {
                isLast = true;
            }
            receivableInItemVO = receivableDOList.get(i);
            BigDecimal writeOffBaseAmount = Optional.ofNullable(receivableInItemVO.getWriteOffBaseAmount()).orElse(BigDecimal.ZERO);
            BigDecimal notWriteOffAmount = receivableInItemVO.getBaseAmount().subtract(writeOffBaseAmount);
            if (notWriteOffAmount.compareTo(BigDecimal.ZERO) <= 0) {
                continue;
            }
            BigDecimal currentWriteOffAmount;
            if (isLast) {
                currentWriteOffAmount = writeOffAmount;
            } else {
                currentWriteOffAmount = writeOffAmount.min(notWriteOffAmount);
            }
            writeOffAmount = writeOffAmount.subtract(currentWriteOffAmount);
            ReceivableWriteOffReqVO writeOffReqVO = buildWriteOffReqVO(receivableInItemVO, currentWriteOffAmount);
            receivableWriteOffReqVOS.add(writeOffReqVO);
            if (isWriteOff) {
                updateReceivableWriteOffAmount(receivableInItemVO, currentWriteOffAmount);
            }
            if (writeOffAmount.compareTo(BigDecimal.ZERO) <= 0) {
                break;
            }
        }
        return receivableWriteOffReqVOS;
    }

    private ReceivableWriteOffReqVO buildWriteOffReqVO(ReceivableDO receivableInItemVO, BigDecimal writeOffAmount) {
        BigDecimal exchangeRate = receivableInItemVO.getExchangeRate();
        BigDecimal receivableWriteOffAmount = writeOffAmount.divide(exchangeRate, 6, RoundingMode.HALF_UP);
        return ReceivableWriteOffReqVO.builder()
                .receivableId(receivableInItemVO.getId())
                .writeOffAmount(writeOffAmount)
                .orderId(receivableInItemVO.getOrderId())
                .receivableWriteOffAmount(receivableWriteOffAmount)
                .build();
    }

    private void updateReceivableWriteOffAmount(ReceivableDO receivableInItemVO, BigDecimal currentWriteOffAmount) {
        if (receivableInItemVO.getWriteOffBaseAmount() == null) {
            receivableInItemVO.setWriteOffBaseAmount(BigDecimal.ZERO);
        }
        BigDecimal writeOffBaseAmount = receivableInItemVO.getWriteOffBaseAmount().add(currentWriteOffAmount);
        BigDecimal writeOffAmountActual = Optional.ofNullable(receivableInItemVO.getWriteOffAmount()).orElse(BigDecimal.ZERO);
        BigDecimal newWriteOffAmount = writeOffAmountActual.add(currentWriteOffAmount.divide(receivableInItemVO.getExchangeRate(), 6, RoundingMode.HALF_UP));
        receivableInItemVO.setWriteOffBaseAmount(writeOffBaseAmount);
        receivableInItemVO.setWriteOffAmount(newWriteOffAmount);
    }


    private void createWriteOffRecord(ReceivableWriteOffReqVO receivableWriteOffReqVO, ReceiptItemDO receiptItem) {
        BankAccountDTO account = bankApi.getBankAccountByAccountId(receiptItem.getAccountId());
        Integer incomeBelong = account.getBaIncomeBelong();
        ReceivableWriteOffRecordDO writeOffRecord = ReceivableWriteOffRecordDO.builder()
                .receivableId(receivableWriteOffReqVO.getReceivableId())
                .writeOffAmount(receivableWriteOffReqVO.getWriteOffAmount())
                .receiptItemId(receiptItem.getId())
                .incomeBelong(incomeBelong)
                .receivableWriteOffAmount(receivableWriteOffReqVO.getReceivableWriteOffAmount())
                .orderId(receivableWriteOffReqVO.getOrderId())
                .build();
        receivableWriteOffRecordMapper.insert(writeOffRecord);
    }
}
